# Implements an aliens-like game

from turtle import Turtle,mainloop
import random
import math

# ------------------------------------------
# Laser cannon (inherits from Turtle superclass)
class LaserCannon(Turtle):
    
    # constructor
    def __init__(self,xmin,xmax,ymin,ymax):
        super().__init__()
        
        # setup world
        self.screen = self.getscreen()
        self.screen.bgcolor('light green')
        self.screen.setworldcoordinates(xmin,ymin,xmax,ymax)
        
        # specify key/mouse bindings (which functions get called)
        self.screen.onclick(self.aim,1)
        self.screen.onkey(self.shoot,"s")
        self.screen.onkey(self.quit,'q')

    # point laser
    def aim(self,x,y):
        heading = self.towards(x,y)
        self.setheading(heading)

    # pew pew
    def shoot(self):
        Bomb(self.heading(),5)
        
    # on finish
    def quit(self):
        self.screen.bye()
        
# ------------------------------------------
# Bounded Turtle (inherits from Turtle superclass)  
class BoundedTurtle(Turtle):
    
    # constructor
    def __init__(self, speed, xmin=-200,xmax=200,ymin=0,ymax=400):
        super().__init__()
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.speed = speed

    # check whether it is out of bounds
    def outOfBounds(self):
        xpos,ypos = self.position()
        out = False
        if xpos < self.xmin or xpos > self.xmax:
            out = True
        if ypos < self.ymin or ypos > self.ymax:
            out = True
        return out
    
    # move to (x,y) coordinate or remove if out of bounds
    def move(self):
        self.forward(self.speed)
        if self.outOfBounds():
            self.remove()
        else:
            self.getscreen().ontimer(self.move,200)
    
    # hide
    def remove(self):
        self.hideturtle()

# ------------------------------------------
# Bomb (inherits from BoundedTurtle superclass)   
class Bomb(BoundedTurtle):
    
    # constructor
    def __init__(self, initHeading,speed):
        super().__init__(speed)
        self.initHeading = initHeading
        self.resizemode('user')
        self.color('red','red')
        self.shape('circle')
        self.setheading(initHeading)
        self.up()
        self.turtlesize(.25)
        self.getscreen().ontimer(self.move,100)
    
    # move forward
    def move(self):
        exploded = False
        self.forward(self.speed)
        
        # explode any nearby aliens
        for i in Alien.getAliens():
            if self.distance(i) < 5:
                i.remove()
                exploded = True
        
        # check if bomb already exploded
        if self.outOfBounds() or exploded:
            self.remove()
        else:
            self.getscreen().ontimer(self.move,100)

    # compute distance to some other item
    def distance(self,other):
        p1 = self.position()
        p2 = other.position()
        a = p1[0]-p2[0]
        b = p1[1]-p2[1]
        dist = math.sqrt(a**2 + b**2)
        return dist

# ------------------------------------------
# Alien (inherits from BoundedTurtle superclass)
class Alien(BoundedTurtle):
    alienList = []

    # return all aliens that are alive
    # note: a static method is bound to a class rather than an object (class instance)
    @staticmethod
    def getAliens():
        return [x for x in Alien.alienList if x.alive]  
    
    # constructor
    def __init__(self,speed,xmin,xmax,ymin,ymax):
        super().__init__(speed,xmin,xmax,ymin,ymax)
        self.getscreen().tracer(0)
        self.up()
        if 'PurpleAlien.gif' not in self.getscreen().getshapes():   
            self.getscreen().addshape('PurpleAlien.gif')            
        self.shape('PurpleAlien.gif')
        self.goto(random.randint(xmin-1,xmax-1),ymax-20)
        self.setheading(random.randint(250,290))
        self.getscreen().tracer(1)
        Alien.alientList = [x for x in Alien.alienList if x.alive]
        Alien.alienList.append(self)
        self.alive = True
        self.getscreen().ontimer(self.move,200)
    
    # remove alien
    def remove(self):
        self.alive = False
        self.hideturtle()

# ------------------------------------------
# AlienInvaders class
class AlienInvaders:
    
    # constructor
    def __init__(self,xmin,xmax,ymin,ymax):
        super().__init__()
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

    # play the game - setup lasercanon and window
    def play(self):
        self.mainWin = LaserCannon(self.xmin,self.xmax,self.ymin,self.ymax).getscreen()
        self.mainWin.ontimer(self.addAlien,1000)
        self.mainWin.listen()
        mainloop()

    # add a new alien
    def addAlien(self):
        if len(Alien.getAliens()) < 7:
            Alien(1,self.xmin,self.xmax,self.ymin,self.ymax)
        self.mainWin.ontimer(self.addAlien,1000)

# main point of entry for program
game = AlienInvaders(-200,200,0,400)
game.play()
